/*======================================================================================

  Tffind.h release version 1.0
  Coded by Matt Weirauch (weirauch@soe.ucsc.edu)
  Aug 11, 2003

  For documentation and source code, see  http://globin.cse.psu.edu/ or
					  http://bio.cse.psu.edu/
  For command line options, type "tffind"

  ======================================================================================*/

#ifndef TFFIND
#define TFFIND

#include "util.h"
#include "seq.h"
#include "read_malign.h"
#include "bb_util.h"


#define MENU_BORDER "\n\n==============================================================" \
		    "=====\n\n"
#define CHOICE_PROMPT "Please enter your menu choice.   "
#define SYNTAX  "ALIGNMENT_FILE MATRIX_FILE_DATABASE > OUTPUT_FILE\n(interactive mode)" \
		"\n\ntffind ALIGNMENT_FILE MATRIX_FILE_DATABASE [-n TFBS_NAME]" \
		"[-p TFBS_CONSENSUS_PATTERN]" \
		"\n  [-a TFBS_ACCESS_NUMBER] [-i TFBS_TRANSFAC_ID] [-s \'SPACER\']" \
		"\n  [-list] [-db MATRIX_FILE_DATABASE] [-min MIN_SEQS] [-cut CUTOFF]" \
		"\n  [-range FROM TO] [-lookin SEQS_TO_LOOK_IN]" \
		"\n  [-exons g|c|e WHICH_SEQ EXONS_FILENAME] [-reltop]" \
		"\n > OUTPUT_FILE\n(batch mode)\n\n"

#define SCREEN stderr                   	/* output intended to go to the screen */
#define STR_SIZE (128 * sizeof (char))  	/* maximum string size */
#define BUFSIZE 10000                     	/* maximum buffer size */
#define MAX_SUBMAT_LEN 100 			/* maximum length of motif */
#define REVERSE 0                       	/* reverse strand */
#define FORWARD 1                               /* forward strand */
#define NAME 0                          	/* searching by name */
#define PAT 1                           	/* or searching by pattern */
#define AC 2					/* or searching by access number */
#define ID 3					/* or searching by i.d. */
#define NOKILL 0				/* do not terminate program on error */
#define KILL 1					/* terminate program on error */
#define WARN 2					/* warning instead of error */
#define rowA 0                          	/* array indices for each nucleotide */
#define rowC 1
#define rowG 2
#define rowT 3
#define BBTAG "#:bb"				/* tag at top of block builder output */
#define BBLEN 4					/* length of BBTAG */
#define DEFAULT_CUTOFF_VAL 0.85			/* def for cutoff used to compute score */

char *err_string;				/* global string for err msg formatting */


/* One item in the linked list of subpatterns.
   An item can represent either a motif or a spacer. */
struct ITEM
{
    int length;                         /* length of the motif, set to -1 if a spacer */
    int compl;                          /* look on reverse strand? */
    int first;                          /* spacer values, set to 0 if a motif */
    int second;
    float max_score;                    /* maximum possible score */
    float min_score;			/* minimum possible score */
    float threshold;                    /* minimum required score */
    float **submat;                     /* the substitution matrix */
    struct ITEM *next;                  /* link to next member of linked list */
};


/* One match to the given pattern. */
struct MATCH
{
    int *positions;                     /* positions of the motifs in the match, used to
					   check if matches on different seqs in the
					   alignment are the same match */
    int *which_seqs;                    /* which sequences match? */
    int *startpt;                       /* starting pt on each seq of match for output */
    int *endpt;                         /* ending pt on each seq of match for output */
    int *block_nbrs;                    /* which blocks is the match located in? */
    int num_seqs;                       /* how many sequences match? */
    int match_len;                      /* how long is the match? */
    int gap_in_top;                     /* does match align to a gap in the top seq? */
    float *pct;                         /* the lowest cutoff val this match would count */
    struct MATCH *next;                 /* link to next member of linked list */
};



static void add_motif_to_list( int compl, char *string, char **pat_string, float cutoff,
                               FILE *fp, char file_type, struct ITEM **item_head,
			       int *max_pat_len );
static void add_spacer( int first, int second, char **pat_string, char *long_string,
                        struct ITEM **item_head, int *max_pat_len );
static void add_to_matches( int num_motifs, int which_seq, int nrow,
                            struct MATCH **match_head, int match_len, float pct,
			    int *block_nbrs, int startpt, int endpt, int gap_in_top,
			    int report_to_top, int *bp_positions, int cur_dir );
static void add_to_match_list( struct MATCH *new_match, struct MATCH **match_head, 
                               int num_motifs, int nrow, int report_to_top, int cur_dir );
static void adjust_from_and_to( int *from, int *to, int *col2pos, int ncol, int cur_dir,
				char align_type, int max_pat_len );
static int can_parse_spacer( char* spacer, int *first, int *second );
static void change_database( char *current_mat_file, char *file_type, FILE **fp,
                             int kill_type );
static void check_boundaries( int *start, int *end, int lowest_pos, int highest_pos );
static void check_exon_positions( int int1, int int2, char *buf );
static void delete_last_node( struct ITEM **item_head, int *num_motifs );
static void delete_last_pat( char **pat_string );
static char determine_align_type( char *align_file );
static void fill_in_match_info( struct MATCH *new_match, int num_motifs, int which_seq,
                                int match_len, float pct, int nrow, int *block_nbrs, 
				int startpt, int endpt, int gap_in_first,
				int *bp_positions );
static void get_avg_pcts( float **pcts_arr, int tot_matches, int num_motifs_in_pat );
static void get_lowest_pcts( float **pcts_arr, int tot_matches, int num_motifs_in_pat );
static void get_position_in_seq( char **A, int ncol, int which_seq, int **col2pos );
static int get_data_from_file( FILE *fp, char *string, int type, struct ITEM **item_head,
                               char **pat_string, float cutoff, char file_type,
                               int *num_motifs, int *max_pat_len );
static void get_endpoints( int j, int dir, int match_len, int ncol, int *col2pos,
                           int *start, int *end, int *gap_in_first, char align_type,
			   Block_ptr block, Block_ptr block_head, Contig_ptr contig,
			   int which_row, int report_to_top );
static void get_ends_b_rev( int pos, int match_len, int *start, int *end,
			    int *gap_in_first, Block_ptr block, Block_ptr block_head,
			    Contig_ptr contig, int which_row, int report_to_top );
static void get_ends_b_for( int pos, int match_len, int *start, int *end,
                            int *gap_in_first, Block_ptr block, Block_ptr block_head,
                            Contig_ptr contig, int which_row, int report_to_top );
static void get_ends_m_rev( int pos, int match_len, int ncol, int *col2pos, int *start, 
			    int *end, int *gap_in_first, int report_to_top );
static void get_ends_m_for( int pos, int match_len, int ncol, int *col2pos, int *start,
			    int *end, int *gap_in_first, int report_to_top );
static int get_exon_info( int **begin, int **end, int **dir, int *exon_seq, int nrow,
			  char **exon_file );
static int get_exons( FILE *fp, int **begin, int **end, int **dir, int mask_type );
static float get_max_score( struct ITEM *item );
static float get_min_score( struct ITEM *item );
static char get_menu_choice( );
static int get_next_window( int *start, int *end, struct ITEM **item_step );
static void get_range( int *from, int *to, int ncol, int highest_pos );
static float get_score( struct ITEM *step, char seq_char, int pos );
static char guess_matrix_type( char *buf );
static float motif_score( char *seq, struct ITEM *step, int cp, int ncol,
			  Block_ptr b_ptr, Block_ptr block_head, Contig_ptr contig,
			  int which_row, int cur_dir );
static void move_to_adjacent_block( Block_ptr *block, int *ncol, Block_ptr block_head,
				    Contig_ptr contig, int *which_row, int cur_dir );
static int output_match( char **A, struct MATCH *match_step, int nrow, int cur_dir,
			 int old_from, int old_to, char align_type, Block_ptr block,
		 	 Block_ptr block_head, Contig_ptr contig, int filter_exons,
			 int *beg, int *end, int *dir, int num_to_mask,
			 int *num_filtered_exons, int num_motifs, int exon_seq );
static void output_matches( struct MATCH *match_head, int min_matches, int filter_exons,
			    int *num_filtered_exons, int *tot_matches, char **A, int nrow,
			    int cur_dir, int num_to_mask, int *beg, int *end, int *dir,
    			    int old_from, int old_to, char align_type,
			    Block_ptr block_head, Contig_ptr contig, int num_motifs,
			    int exon_seq );
static void print_header( char *align_file, int from, int to, char *pat_string, int nrow,
                          int ncol, int filter_exons, int exon_seq, int *look_in_seq,
			  int min_matches, char align_type, int report_to_top, 
			  int num_motifs, char *exon_file );
static void print_main_menu( char *pat_string, char *mat_file, int nrow, float cutoff,
                             int min_matches, int from, int to, int *look_in_seq,
                             char *align_file, int filter_exons, 
			     int ncol, int report_to_top );
static void print_matches( struct MATCH *match_head, int num_motifs, int nrow );
static void print_trailer( int tot_matches, int filter_exons, int num_filtered_exons,
			   int nrow, char **seq_names, int report_to_top,
			   char align_type );
static void process_cmnd_line_options( char **pat_string, struct ITEM **item_head,
                                       int *num_motifs, int *from, int *to,
                                       int *look_in_seq, int *filter_exons, int nrow,
                                       int ncol, int *min_matches, int *report_to_top,
				       int **beg, int **end, int **dir, int *exon_seq,
				       int *num_to_mask, int argc, char **argv,
				       int *max_pat_len, char *current_mat_file,
				       int highest_pos, char **exon_file );
static void process_main_menu_choices( char **pat_string, struct ITEM **item_head, 
                                       int *num_motifs, int *from, int *to,
                                       char *align_file,int *look_in_seq,
                                       int *filter_exons, int nrow, int ncol,
				       int *min_matches, int *report_to_top, int **beg,
				       int **end, int **dir, int *exon_seq,
				       int *num_to_mask, int *max_pat_len, 
				       char *current_mat_file, int highest_pos,
				       char **exon_file );
static void read_matrix( FILE *fp, struct ITEM *new_item, char file_type);
static int record_matches_in_window( int start, int end, int offset, int cur_match,
				     struct ITEM **item_step, char *seq, int ncol,
				     int ***match_arr, int *which_match, int which_motif,
				     int num_motifs_in_par, float ***pcts_arr, 
				     Block_ptr b_ptr, int ***block_nbr_arr, 
				     Block_ptr block_head, Contig_ptr contig,
				     int which_row, int cur_dir, int ***bp_arr, 
				     int prev_motif_len );
static void record_match( int ***match_arr, int *which_match, int which_motif,
                          int cur_pos, float score, float max_score, float ***pcts_arr,
			  int num_motifs_in_pat, int cur_match, int ***block_nbr_arr,
			  int block_nbr, int bp, int ***bp_arr, float min_score );
static void record_pattern_matches( struct MATCH **match_head, struct ITEM *item_head,
                                    char *seq, int nrow, int ncol, int i, int j,
				    int num_motifs_in_pat, int cur_dir, char align_type,
				    Block_ptr b_ptr,Block_ptr block_head,
				    Contig_ptr contig, int **col2pos, int report_to_top );
static void remove_trailing_spacers( struct ITEM **item_head, char **pat_string );
static int same_match( struct MATCH *match_step, int *positions, int num_motifs,
                       int *block_nbrs );
static void scan_matrix_file( FILE *fp, char *current_mat_file, char file_type);
static void search_for_pattern( char **A, int ncol, int nrow, struct ITEM *item_head,
                                int num_motifs, int from, int to, int cur_dir,
				int *look_in_seq, struct MATCH **match_head,
				char align_type, Block_ptr block_head, Contig_ptr contig,
				int ** col2pos, int report_to_top, int min_matches );
static int too_far_in_match_list( struct MATCH *match_step, struct MATCH *new_match,
                         	  int num_motifs, int nrow, int report_to_top,
				  int cur_dir );
static void update_exon_info( int *begin, int *end, int *dir, int *i, int num1, int num2,
                              int cur_dir );
static int valid_match( int *match, int num_motifs_in_pat );
static int within_exon( int startpt, int endpt, int *beg, int *end, int *dir, int cur_dir,
                        int num_to_mask );

#endif
/*======================================================================================*/
